# This file is part of the graph-includes package
#
# (c) 2005 Yann Dirson <ydirson@altern.org>
# Distributed under version 2 of the GNU GPL.

package graphincludes::project;

sub new {
  my $class = shift;
  my ($prefixstrip, @files) = @_;
  my $self  = {};

  $self->{FILES}   = \@files;
  $self->{PFXSTRIP}= $prefixstrip;
  $self->{TREE}    = undef;
  $self->{NODEIDS} = undef;
  $self->{DEPS}    = undef;
  $self->{SPECIALEDGES} = {};

  bless ($self, $class);

  return $self;
}

sub init {
  my $self = shift;

  # store as a tree while keeping recording node id's
  $self->{NODEIDS} = $self->nodeids($self->{TREE}, @{$self->{FILES}});
  $self->{DEPS} = $self->getdeps();
}

sub nodeids {
  my $self = shift;
  my ($tree, @files) = @_;
  my $i=0;
  my %nodeid;
  foreach my $file (@files) {
    my $label = $self->filelabel($file);
    ## FIXME: why is return/assignment necessary !?
    #$tree = $self->insert ($label, $tree);
    if (!defined $nodeid{$label}) {
      $nodeid{$label} = "f" . ++$i;
      $nodeid{$label} .= '-' . $label if $main::debug;
    }
  }
  return \%nodeid;
}

sub reduce {
  my $self = shift;
  my %reduceddeps = %{$self->{DEPS}};

  our $dropcount=0;
  sub suppress($$$@) {
    my ($list,			# childlist to cleanup
	$node,			# node to consider this time
	$considered,		# graph nodes already seen, not to reconsider
	@context)		# current path, for printing only
      = @_;

    # Do not consider $node twice, prevent looping on circular deps.
    # We must take care of the special case of the child that led us to
    # the current node, or we would have to do special things to $gchild
    return if $node eq $context[1] or grep { $node eq $_ } (@$considered);
    push @$considered, $node;

    # remove $node from $list
    my @newlist = @$list;
    foreach my $child (@$list) {
      if ($child eq $node) {
	if ($main::showdropped) {
	  $self->{SPECIALEDGES}->{$considered->[0]}->{$child} = 'color="#FFCCCC",constraint=false';
	} elsif (grep { $self->{NODEIDS}->{$_} eq $considered->[0] } @main::focus) {
	  $self->{SPECIALEDGES}->{$considered->[0]}->{$child} = 'color="#FFCCCC"';
	} else {
	  $dropcount++;
	  # FIXME: make more efficient ?
	  @newlist = map { $_ eq $node ? () : $_ } @newlist;
	}
	print STDERR "    --$child (", join (',', @context), ")\n" if $main::debug;
	last;
      }
    }
    splice (@$list, 0, 0+@$list, @newlist);

    # remove $node's children
    if (defined $reduceddeps{$node}) {
      foreach my $child (@{$reduceddeps{$node}}) {
	if ($main::debug) {
	  foreach (@context) {
	    print STDERR " ";
	  }
	  print STDERR "$child\n";
	}
	suppress ($list, $child, $considered, (@context, $child));
      }
    }
  }

  print STDERR "Doing transitive reduction " if $main::verbose;
  foreach my $node (keys %reduceddeps) {
    print STDERR '.' if $main::verbose;
    print STDERR "node $node\n" if $main::debug;
    if (defined $reduceddeps{$node}) {
      my @newdeps = @{$reduceddeps{$node}};
      my @considered = ($node);
      foreach my $child (@{$reduceddeps{$node}}) {
	# do not explore children already removed, or some circles cause lost edges
	next unless grep { $child eq $_ } @newdeps;
	print STDERR " child $child\n" if $main::debug;
	if ($child eq $node) {
	  # remove self-deps induced by trivial #include
	  print STDERR "  --$child\n" if $main::debug;
	  @newdeps = map { $_ eq $node ? () : $_ } @newdeps;
	} else {
	  if (defined $reduceddeps{$child}) {
	    foreach my $gchild (@{$reduceddeps{$child}}) {
	      if ($gchild ne $node and $gchild ne $child) { # XXX
		print STDERR "  gchild $gchild\n" if $main::debug;
		suppress (\@newdeps, $gchild, \@considered, ($node, $child, $gchild));
	      }
	    }
	  }
	}
      }
      $reduceddeps{$node} = \@newdeps;
    }
  }
  print STDERR " $dropcount cleared.\n" if $main::verbose;

  print STDERR "Verifying validity of transitive reduction " if $main::verbose;
  sub has_path {
    my ($graph, $from, $to, @seen) = @_;
    return 1 if $from eq $to;
    return 0 if grep { $_ eq $from } @seen;
    return 1 if grep { $_ eq $to } @{$graph->{$from}}; # superfluous ?
    foreach my $child (@{$graph->{$from}}) {
      return 1 if has_path($graph, $child, $to, (@seen, $from));
    }
    return 0;
  }
  foreach my $node (keys %{$self->{DEPS}}) {
    print STDERR '.' if $main::verbose;
    foreach my $child (@{$self->{DEPS}->{$node}}) {
      if (!has_path(\%reduceddeps, $node, $child)) {
	print STDERR "ERROR: missing edge from $node to $child\n";
      }
    }
  }
  print STDERR " done.\n" if $main::verbose;
  $self->{DEPS} = \%reduceddeps;
}

sub special_edge {
  my $self = shift;
  my ($src, $dst) = @_;

  my $attrs = $self->{SPECIALEDGES}->{$src}->{$dst};

  if (defined $attrs) {
    return "$attrs";
  } else {
    return undef;
  }
}

1;
